#ifndef CoCoA_DenseUPolyRing_H
#define CoCoA_DenseUPolyRing_H

//   Copyright (c)  2007  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Header file for the abstract class DenseUPolyRingBase etc.


#include "CoCoA/error.H"
#include "CoCoA/ideal.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/RingHom.H"

#include <cstddef>
//using std::size_t;
#include <iosfwd>
//using std::ostream;
#include <vector>
//using std::vector;
#include <memory>
// using std::auto_ptr;


namespace CoCoA
{

  class symbol;  // fwd declaration -- defined in symbol.H

  class DenseUPolyRingBase;

  /*-----------------------------------------------------------------*/
  /** \include DenseUPolyRing.txt  */
  /*-----------------------------------------------------------------*/
  class DenseUPolyRing: public PolyRing
  {
  public:
    explicit DenseUPolyRing(const DenseUPolyRingBase* RingPtr);
    // Default copy ctor works fine.
    // Default dtor works fine.
  private: // disable assignment
    DenseUPolyRing& operator=(const DenseUPolyRing& rhs); ///< NEVER DEFINED -- assignment disabled
  public:
    const DenseUPolyRingBase* operator->() const; ///< allow member fns to be called
  };


  class DenseUPolyRingBase: public PolyRingBase
  {
  public:
    ///@name Member functions every concrete DenseUPolyRing implementation must have in addition to those of PolyRingBase.
    //@{
    virtual const symbol& myIndetName() const = 0;
    virtual std::size_t mySize(ConstRawPtr rawf) const = 0;            ///< returns (memory) size of f
    virtual std::size_t myDegPlus1(ConstRawPtr rawf) const = 0;        ///< returns deg of f +1 (0 for zero poly)
    virtual ConstRefRingElem myCoeff(ConstRawPtr rawf, std::size_t d) const = 0;
    virtual void myAddMul(RawPtr rawf, ConstRawPtr rawc, std::size_t d, ConstRawPtr rawg) const = 0; ///<  f += c*indet^d*g
    virtual void myMulByXExp(RawPtr rawf, unsigned long d) const = 0;
    virtual void myMulBy1MinusXExp(RawPtr rawf, unsigned long d) const = 0;
    //@}

    ///@name   Functions for creating/building polynomials
    //@{
    virtual RingElem myMonomial(ConstRawPtr rawc, unsigned long d) const; // c is element of CoeffRing!!
    virtual void myAssignZeroCoeff(RawPtr rawf, std::size_t d) const = 0; ///< f_d = 0, no check on size
    virtual void myAssignNonZeroCoeff(RawPtr rawf, ConstRawPtr rawc, std::size_t d) const = 0; ///< f_d = c, no check on size
    virtual void myResize(RawPtr rawf, std::size_t NewSize) const = 0;
    virtual void myResetDeg(RawPtr rawf) const = 0;
    //@}

    ///@name functions inherited from PolyRing with a single implementation for all DenseUPolyRing implementations
    //@{
    virtual std::size_t myNumIndets() const {return 1;}
    virtual void myIndetPower(RawPtr rawf, std::size_t var, unsigned int exp) const;
    virtual std::size_t myNumTerms(ConstRawPtr rawf) const;
    virtual bool myIsConstant(ConstRawPtr rawf) const;                  ///< true iff rawf == c (possibly 0)
    virtual bool myIsIndet(std::size_t& IndetIndex, ConstRawPtr rawf) const;///< true iff rawf is x[IndetIndex]
    virtual bool myIsMonomial(ConstRawPtr rawf) const;                  ///< is it useful for univariate?
    virtual std::size_t myStdDeg(ConstRawPtr rawf) const;              ///< standard degree of f, error for zero poly
    virtual long myLog(ConstRawPtr rawf, size_t index) const;          ///< max exponent of index (same as myStdDeg)
    virtual ConstRefRingElem myLC(ConstRawPtr rawf) const;
    virtual void myContent(RawPtr rawcontent, ConstRawPtr rawf) const;
    virtual void myRemoveBigContent(RawPtr rawf) const;
    virtual bool myIsValid(ConstRawPtr rawf) const;                    ///< true iff f is a valid representation
    //    virtual bool myIsHomogeneous(ConstRawPtr rawf) const;                  ///< true iff f is homogeneous
    //@}

    ///@name functions inherited from ring with a single implementation for all DenseUPolyRing implementations
    //@{
    virtual void myMul(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; ///< lhs = x*y
    virtual void mySymbols(std::vector<symbol>& SymList) const;         ///< append symbols in ring to SymList
    virtual void myOutput(std::ostream& out, ConstRawPtr rawx) const;   ///< out << x
    virtual bool myIsZero(ConstRawPtr rawx) const;                      ///< x == 0
    virtual bool myIsPrintAtom(ConstRawPtr rawx) const;                 ///< <indet>^n may be printed without parentheses
    virtual void myOutput(OpenMathOutput& OMOut, ConstRawPtr rawx) const;///< OMOut << x
    virtual bool myIsOne(ConstRawPtr rawf) const;                       ///< true iff rawf == 1
    virtual bool myIsMinusOne(ConstRawPtr rawf) const;                  ///< true iff rawf == -1
    bool myIsEqual(ConstRawPtr rawx, ConstRawPtr rawy) const;           ///< true iff rawx == rawy
    virtual void myGcd(RawPtr rawlhs, ConstRawPtr rawx, ConstRawPtr rawy) const; ///< lhs = gcd(x,y) if GCDDomain;
    virtual bool myIsInteger(ZZ& N, ConstRawPtr rawx) const;               ///< true iff x is integer
    virtual bool myIsRational(ZZ& N, ZZ& D, ConstRawPtr rawx) const;       ///< true iff x is rational
    virtual bool myIsInvertible(ConstRawPtr rawx) const;                   ///< true iff x is invertible constant
    virtual ideal myIdealCtor(const std::vector<RingElem>& gens) const;
    virtual void myPowerSmallExp(RawPtr rawlhs, ConstRawPtr rawx, unsigned long n) const;// lhs = x^n, n>1, x not -1,0,1
    //@}

    ///@name  Functions for homomorphisms
    //@{
    virtual RingHom myHomCtor(const ring& codomain, const RingHom& CoeffHom, const std::vector<RingElem>& IndetImages) const;
    //    RingHom myHomCtor(const ring& codomain, const RingHom& CoeffHom, ConstRefRingElem IndetImage) const;
    RingHom myCompose(const RingHom& phi, const RingHom& theta) const;
    //@}

  protected: // Ideal class for this type of ring.

    class IdealImpl: public IdealBase
    {
    public:
      IdealImpl(const DenseUPolyRing& P, const std::vector<RingElem>& gens);
      // default copy ctor is OK
      virtual IdealBase* clone() const;
//???    virtual void swap(ideal& other);

      virtual const DenseUPolyRing& myAmbientRing() const;
      virtual bool IamZero() const;
      virtual bool IamOne() const;
      virtual void myReduceMod(RingElemRawPtr rawx) const; // r elem of R, where I is ideal in R
      virtual bool contains(RingElemConstRawPtr rawx) const;
      virtual void add(const ideal&);
      virtual void intersect(const ideal&);
      virtual void colon(const ideal&);
      virtual void myDivMod(RingElemRawPtr rawlhs, RingElemConstRawPtr rawnum, RingElemConstRawPtr rawden) const; // lhs = num/den modulo the ideal  (lhs = 0 if quotient does not exist)

      virtual const std::vector<RingElem>& myGens() const; // gens as specified by user
      virtual const std::vector<RingElem>& myTidyGens() const; // tidier set of gens
    protected:
      virtual void myMaximalTest() const;
      virtual void myPrimeTest() const;

    protected: // Data members of DenseUPolyRingBase::IdealImpl
      DenseUPolyRing myP;
      std::vector<RingElem> myGensValue;
      //      mutable bool myGBasisIsValid;
      //      mutable std::vector<RingElem> myGBasisValue;
    protected: // Auxiliary functions.
      //      std::vector<RingElem>& myGBasis() const;
      static const DenseUPolyRingBase::IdealImpl* GetPtr(const ideal& J);
    };  // class IdealImpl

  protected:  // General homomorphism class for this type of ring.

    class HomImpl: public RingHomBase
    {
    public:
      HomImpl(const DenseUPolyRing& domain, const ring& codomain, const RingHom& CoeffHom, ConstRefRingElem IndetImage);
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    private:
      virtual void myOutputSelfDetails(std::ostream& out) const;

    private: // Data members of DenseUPolyRingBase::HomImpl
      const RingHom myCoeffHom;
      ConstRefRingElem myIndetImage;
    };  // class HomImpl

  protected:  // Special homomorphism class for this type of ring.

    class CoeffEmbeddingHomImpl: public RingHomEmbeddingBase
    {
    public:
      CoeffEmbeddingHomImpl(const DenseUPolyRing& P);
      virtual void myApply(RingElemRawPtr rawimage, RingElemConstRawPtr rawarg) const;
    };  // class CoeffEmbeddingHomImpl

  };


  bool IsDenseUPolyRing(const ring& R);                         // inline
  DenseUPolyRing AsDenseUPolyRing(const ring& R);                // inline
  bool IsGoodIndetName(const ring& CoeffRing, const symbol& IndetName);

  ///@name Pseudo constructors for RingElems owned by DenseUPolyRing
  //@{
  RingElem monomial(const DenseUPolyRing& P, ConstRefRingElem c, unsigned long exp); ///< c*indet^exp
  RingElem monomial(const DenseUPolyRing& P, const ZZ& N, unsigned long exp);        ///< N*indet^exp
  RingElem monomial(const DenseUPolyRing& P, long n, unsigned long exp);             ///< n*indet^exp
  //@}

  ///@name Special functions on RingElems owned by DenseUPolyRing
  //@{
  inline ConstRefRingElem coeff(ConstRefRingElem f, std::size_t d)
  { return AsDenseUPolyRing(owner(f))->myCoeff(raw(f), d); }
  //@}

  //----------------------------------------------------------------------
  // Inline functions

  inline bool IsDenseUPolyRing(const ring& R)
  {
    return dynamic_cast<const DenseUPolyRingBase*>(R.myRawPtr()) != 0;
  }


  inline DenseUPolyRing AsDenseUPolyRing(const ring& R)
  {
    if (!IsDenseUPolyRing(R))
      CoCoA_ERROR(ERR::NotDenseUPolyRing, "AsDenseUPolyRing(ring)");
    return DenseUPolyRing(static_cast<const DenseUPolyRingBase*>(R.myRawPtr()));
  }


  inline DenseUPolyRing::DenseUPolyRing(const DenseUPolyRingBase* RingPtr):
    PolyRing(RingPtr)
  {}


  inline const DenseUPolyRingBase* DenseUPolyRing::operator->() const
  {
    return static_cast<const DenseUPolyRingBase*>(ring::operator->());
  }
  

} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/DenseUPolyRing.H,v 1.4 2007/10/30 17:14:12 abbott Exp $
// $Log: DenseUPolyRing.H,v $
// Revision 1.4  2007/10/30 17:14:12  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/10/15 12:38:52  bigatti
// -- fixed comment
//
// Revision 1.2  2007/10/10 14:02:37  bigatti
// -- added myMulBy1MinusXExp
// -- fixed a few little bugs
//
// Revision 1.1  2007/10/05 15:28:56  bigatti
// -- added abstract class DenseUPolyRing for representing dense
//    univariate polynomials
// -- added concrete class RingDenseUPolyClean, the cleanest
//    implementation
// -- just for testing, still horribly inefficient and incomplete
//

#endif
