//   Copyright (c)  2005  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/ReductionCog.H"
#include "CoCoA/TmpGPoly.H"
#include "CoCoA/assert.H"
#include "CoCoA/geobucket.H"

#include <algorithm>
using std::max; // for sugar
// #include <vector>
using std::vector;

namespace CoCoA
{

  //-------- headers ----------------------------------------

  // ANNA: where should these go???
  void reduce(ReductionCog& F, const Reductors& v);
  void reduce(ReductionCog& F, const GPoly& g);
  ReductionCog ChooseReductionCogPoly(const GRingInfo& GRI);
  ReductionCog ChooseReductionCogGeobucket(const GRingInfo& GRI);

  //---------------------------------------------------------

  bool ReductorData::myBorelUpdate(ConstRefPPMonoidElem pp, const Reductors& theReductors)
  {
    if (myCount < 100) return false;

    GPoly* f = myGPolyPtr;
    const PPMonoid thePPM(PPM(*f));
    size_t BorelIndetIndex = NumIndets(owner(*f))-1;
    PPMonoidElem quot(thePPM);

    thePPM->myDiv(raw(quot), raw(pp), raw(PP(myLPPwMask)));
    if (quot != IndetPower(thePPM, BorelIndetIndex, log(quot, BorelIndetIndex)))
      return false;

    f->MultiplyByPP(quot);
    myLPPwMask.myAssign(pp);
    IamBorelUpdated = true;
    f->myReduceTail(theReductors);
    return true;
  }


  inline GPoly* FindReducer(const PPWithMask& pm,
                            const unsigned int comp,
                            const Reductors& theReductors)
  {
    for (vector<ReductorData>::iterator it = theReductors.myBorelReductors.begin();
         it != theReductors.myBorelReductors.end();
         ++it)
      if ( comp==it->myComponent && IsDivisibleFast(pm, it->myLPPwMask))
        if (it->IamBorelUpdated || it->myBorelUpdate(PP(pm), theReductors) )
        {
          (it->myCount)++;
          return it->myGPolyPtr;
        }
    for (vector<ReductorData>::const_iterator it = theReductors.myReductors.begin();
         it != theReductors.myReductors.end();
         ++it)
      if (comp==it->myComponent && IsDivisibleFast(pm, it->myLPPwMask)&&!it->IamNotToBeUsed())
      {
        (it->myCount)++;
        return it->myGPolyPtr;
      }
    return 0;
  }


  //---- ReductionCog code --------------------------------------------------

  inline GPoly* FindReducer(const ReductionCog& F, const Reductors& theReductors)
  {
    if ( IsActiveZero(F) ) return 0;
    GRingInfo GRI(theReductors.myGRingInfo());
    const PPWithMask pm(ActiveLPP(F), GRI.myDivMaskRule());
    return FindReducer(pm, GRI.myComponent(PP(pm)), theReductors);
  }


//   inline int FindReducerIndex(const ReductionCog& F, const vector<RingElem>& v)
//   {
//     if ( IsActiveZero(F) ) return -1;
//     return FindReducerIndex(ActiveLPP(F), v);
//   }


  void ReduceActiveLM(ReductionCog& F, const Reductors& v)
  {
    GPoly* g;
    while ( (g = FindReducer(F, v)) != 0 )
      F->myReduce(poly(*g), len(*g));
  }


  // I'm not proud about this style for computing the sugar,
  // but I'm waiting to see how this evolves
  void ReduceActiveLM(ReductionCog& F, const Reductors& v,
                      size_t& SugarPtr,
                      degree& WSugarPtr)
  {
    GPoly* g;
    while ( (g = FindReducer(F, v)) != 0 )
    {
      CoCoA_ASSERT( !IsZero(*g) );
      switch (g->myGRingInfo().myInputAndGrading())
      {
        case AllGraded: break;
	case AffineInputWithGradedRing:
	  if (v.myGRingInfo().IsMyGradingPosPlus())
	  {
	    degree tmp=degree(GradingDim(WSugarPtr));
	    tmp.mySetComponent(0,v.myGRingInfo().TmpStdDeg(ActiveLPP(F)/LPP(*g)));
	    WSugarPtr = max(WSugarPtr,tmp+wsugar(*g));
	  }
	  else
            WSugarPtr = max(WSugarPtr, wdeg(ActiveLPP(F)/LPP(*g))+wsugar(*g));
	break;
	case AllAffine:
	  SugarPtr = max(SugarPtr, v.myGRingInfo().TmpStdDeg(ActiveLPP(F)/LPP(*g))+sugar(*g));
	break;
      }//switch;
      F->myReduce(poly(*g), len(*g));
    }//while
  }//ReduceActiveLM


  void reduce(ReductionCog& F, const Reductors& v)
  {
    ReduceActiveLM(F, v);
    while ( !IsActiveZero(F) )
    {
      F->myMoveToNextLM();
      ReduceActiveLM(F, v);
    }
  }


  void reduce(ReductionCog& F, const Reductors& v,
             size_t& SugarPtr, degree& WSugarPtr)
  {
    ReduceActiveLM(F, v, SugarPtr, WSugarPtr);
    while ( !IsActiveZero(F) )
    {
      F->myMoveToNextLM();
      ReduceActiveLM(F, v, SugarPtr, WSugarPtr);
    }
  }

  void reduce(ReductionCog& F, const GPoly& g,
              size_t& SugarPtr, degree& WSugarPtr)
  {
    if ( IsActiveZero(F) || ActiveLPP(F) < LPP(g) ) return;
    const GRingInfo& GRI(g.myGRingInfo());
    const PPWithMask& PMg(LPPwMask(g));
    const size_t Componentg = GRI.myComponent(PP(PMg));
    PPWithMask PMF(ActiveLPP(F), GRI.myDivMaskRule());
    size_t ComponentF = GRI.myComponent(PP(PMF));
    while ( !IsActiveZero(F) )
    {
      if ( (ComponentF==Componentg && IsDivisibleFast(PMF, PMg) ) )
      {
        switch (GRI.myInputAndGrading())
	{
	  case AllGraded: break;
	  case AffineInputWithGradedRing:
	   if (GRI.IsMyGradingPosPlus())
	   {
	     degree tmp=degree(GradingDim(WSugarPtr));
	     tmp.mySetComponent(0,GRI.TmpStdDeg(ActiveLPP(F)/LPP(g)));
	     WSugarPtr = max(WSugarPtr, tmp+wsugar(g));
	   }
           else
             WSugarPtr = max(WSugarPtr, wdeg(ActiveLPP(F)/LPP(g))+wsugar(g));
	  break;
	  case AllAffine:
           SugarPtr = max(SugarPtr, GRI.TmpStdDeg(ActiveLPP(F)/LPP(g))+sugar(g));
	  break;
	}//switch
        F->myReduce(poly(g), len(g));
      }
      else
        F->myMoveToNextLM();
      if ( IsActiveZero(F) || ActiveLPP(F) < LPP(g) ) return;
      PMF.myAssign(ActiveLPP(F));
      ComponentF = GRI.myComponent(PP(PMF));
    }//while
  }//reduce


  void reduce(ReductionCog& F, const GPoly& g)
  {
    degree ws(wsugar(g));
    size_t s(sugar(g));
    reduce(F, g, s, ws);
  }


  //---- ReductionCog code end ----------------------------------------------


  //-------- ChooseReductionCog... ----------------------------------------

  //ANNA where should these go???
  ReductionCog ChooseReductionCogPoly(const GRingInfo& GRI)
  {
    if ( GRI.myCoeffRingType() == CoeffEncoding::Field )
      return NewRedCogPolyField(GRI.myNewSPR());
    else if ( GRI.myCoeffRingType() == CoeffEncoding::FrFldOfGCDDomain )
      return NewRedCogPolyGCD(GRI.myNewSPR());
    else CoCoA_ERROR("Don't know what to do with these coefficients", "ChooseReductionCog");
    return ReductionCog(0);  // just to keep the compiler quiet
  }


  ReductionCog ChooseReductionCogGeobucket(const GRingInfo& GRI)
  {
    if ( GRI.myCoeffRingType() == CoeffEncoding::Field )
      return NewRedCogGeobucketField(GRI.myNewSPR());
    else if ( GRI.myCoeffRingType() == CoeffEncoding::FrFldOfGCDDomain )
      return NewRedCogGeobucketGCD(GRI.myNewSPR());
    else CoCoA_ERROR("Don't know what to do with these coefficients", "ChooseReductionCog");
    return ReductionCog(0);  // just to keep the compiler quiet
  }


  //-------- GPoly member functions ----------------------------------------

  void GPoly::myPolySetSPoly(const GPoly& f, const GPoly& g)
  {
    myPolyValue = poly(f);
    (owner(f))->myMulByPP(raw(myPolyValue), raw(colon(LPP(g),LPP(f))));
    ReductionCog F = ChooseReductionCogPoly(myGRingInfo());
    F->myAssignReset(myPolyValue, f.myLen);
    F->myReduce(poly(g), len(g));
    F->myRelease(myPolyValue);
    // sugar will be set as the sugar of the originating pair
  }


  void GPoly::myReduceTail(const GPoly& g)
  {
    CoCoA_ASSERT( !IsZero(*this) );
    if ( LPP(*this) < LPP(g) ) return;
    // geobucket because reduce(F, g) might make many reductions by g
    ReductionCog F = ChooseReductionCogGeobucket(myGRingInfo());
    F->myAssignReset(myPolyValue, myLen);
    F->myMoveToNextLM();
    reduce(F, g);
    F->myRelease(myPolyValue);
    myLen = NumTerms(myPoly()); // LPP,Deg,Comp are the same
  }


  void GPoly::myReduceTail(const Reductors& theReductors)
  {
    CoCoA_ASSERT( !IsZero(*this) );
    ReductionCog F = ChooseReductionCogGeobucket(myGRingInfo());
    F->myAssignReset(myPolyValue, myLen);
    F->myMoveToNextLM();
    reduce(F, theReductors, mySugar, myWSugar);
    F->myRelease(myPolyValue);
    myLen = NumTerms(myPoly()); // LPP,Deg,Comp are the same
  }


  void GPoly::myReduce(const Reductors& theReductors)
  {
    if ( IsZero(*this) ) return;
    ReductionCog F = ChooseReductionCogGeobucket(myGRingInfo());
    F->myAssignReset(myPolyValue, myLen); // myPolyValue gets 0
    switch (myGRingInfo().myInputAndGrading())
    {
       case AllGraded:
         reduce(F, theReductors);
       break;
       case AffineInputWithGradedRing:
       case AllAffine:
         reduce(F, theReductors, mySugar, myWSugar); // mySugar, myWSugar are updated
       break;
    }//switch
    F->myRelease(myPolyValue); // myPolyValue gets the value of F
    myUpdateLenLPPLCDegComp(); // myLen, myWDeg, myComponent are updated

    if ( !IsZero(*this) && !IsOne(myLCValue) ) // makes myPolyValue monic
      if ( myGRingInfo().myCoeffRingType()==CoeffEncoding::Field )
        myGRingInfo().myNewSPR()->myDivByCoeff(raw(myPolyValue), raw(myLCValue));
    // if CoeffEncoding::Field myRelease does NOT make poly monic
    // if CoeffEncoding::FrFldOfGCDDomain myRelease makes poly content free
  }




} // end of namespace CoCoA


// RCS Header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/reduce.C,v 1.4 2007/10/30 17:14:06 abbott Exp $
// $Log: reduce.C,v $
// Revision 1.4  2007/10/30 17:14:06  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.3  2007/04/27 14:24:15  bigatti
// -- added commented include for <vector>
//
// Revision 1.2  2007/03/09 18:56:56  bigatti
// -- added Tmp prefix to Groebner related files
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.8  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.7  2007/03/07 17:04:31  cocoa
// -- several changes by M.Caboara: more operations on ideals,
//    exception cleaner, coding conventions, WSugar, dynamic
//
// Revision 1.6  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.5  2006/12/04 13:55:54  cocoa
// -- added: sugar for GradingDim > 0  (called wsugar)
//
// Revision 1.4  2006/10/11 13:33:06  cocoa
// -- rearranged code for sugar in reduce.C
// -- activated sugar in GPair.C
// -- removed operator<< for GPairList (template in io)
//
// Revision 1.3  2006/10/06 16:32:06  cocoa
// -- changed: GPoly::SPoly --> GPoly::myAssignSPoly
// -- changed: Len(const GPoly&) --> len(const GPoly&)
// -- added:   poly(const GPoly&)
// -- added:   GPoly::myUpdateLenLPPDegComp()
// -- in reduce.C added functions for computing sugar during reduction
//
// Revision 1.2  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
